/*----------------------------------------------------------------------*/
/*		hsortp.cpp		hybrid radix / merge text file sort				*/
/*																		*/
/*		Jeff Reid		13APR22	 10:00									*/
/*----------------------------------------------------------------------*/
/*		abbreviations for name prefixes									*/
/*																		*/
/*		a		array of ...											*/
/*		b		unsigned char (byte)									*/
/*		cb		count of bytes in ...									*/
/*		fp		FILE pointer											*/
/*		p		pointer	to ...											*/
/*		dw		unsigned int (dword)									*/
/*		Sorts text files: each record ends in CR(0x0d),	LF(0x0a).		*/
/*																		*/
/*		Records	are	converted into p-strings:							*/
/*			pstring[  0] = c = # bytes in record						*/
/*			pstring[  1] = record data									*/
/*			pstring[c+1] = 0 (terminator)								*/
/*		An array of	pointers to	the	pstrings is	created.				*/
/*		The	array of pointers is sorted.								*/
/*		The	records	are	output using the array of sorted pointers.		*/
/*																		*/
/*		Pointers are split into	NUMBIN bins	based on msb of	records.	*/
/*		Each bin is	sorted.	The	sorted bins	are	combined into one bin.	*/
/*																		*/
/*		Sorting	is done	via	merge sort in 2	steps:						*/
/*		MakeGroups creates the initial groups of pointers.				*/
/*		MergeGroups	merges groups until	there is just 1	group.			*/
/*																		*/
/*		if #define VARSZGRP	is set to 1, then MakeGroups creates		*/
/*		groups based on	ascending or descending	sequences, which is		*/
/*		faster if there	is significant amount of order in the file.		*/
/*		if #define VARSZGRP	is set to 0, then MakeGroups creates		*/
/*		groups of 2	ascending pointers per group.						*/
/*		In either case,	descending sequences are converted to ascending	*/
/*		sequences by MakeGroups.										*/
/*----------------------------------------------------------------------*/
/*		abbreviations for name prefixes									*/
/*																		*/
/*		a		array of ...											*/
/*		b		unsigned char (byte)									*/
/*		cb		count of bytes in ...									*/
/*		fp		FILE pointer											*/
/*		p		pointer	to ...											*/
/*		dw		unsigned int (dword)									*/
/*		qw		unsigned __int64 (qword)								*/
/*		li		LARGE_INTEGER (__int64 structure)						*/
/*----------------------------------------------------------------------*/

/*		get	rid	of warnings	for	fopen */
#define	_CRT_SECURE_NO_WARNINGS	1

#include <iostream>
#include <vector>

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

/*		set	to 1 for variable size groups, 0 for fixed sized groups	*/

#define	VARSZGRP 0

/*		ASCCR is carriage return character */
#define	ASCCR 0x0d

/*		ASCLF is line feed character */
#define	ASCLF 0x0a

typedef	unsigned char		UI8;
typedef	unsigned char *		PUI8;
typedef	unsigned char **	PPUI8;
typedef	unsigned char ***	PPPUI8;
typedef	unsigned int		UI32;
typedef	unsigned int  *		PUI32;
typedef	unsigned int  **	PPUI32;

#define	NUMBIN 161						/* number of bins */

/*----------------------------------------------------------------------*/
/*		data															*/
/*----------------------------------------------------------------------*/

static FILE	*	fpSrc;					/* file	desc for source	*/
static UI32		cbSrc;					/* count for _dos_read */

static FILE	*	fpDst;					/* file	desc dest file */
static UI32		cbDst;					/* count for _dos_write	*/

static UI32		dwNumBytes;				/* # bytes in file */
static UI32		dwNumRec;				/* # of	records	in file	*/
static PUI8		pMem;					/* allocated memory	buffer */
static PUI8		pLast;					/* ptr to last UI8 in file */

static PUI32	pauGroup;				/* ptr to array	of # of	recs */
static PPUI8	papGet;					/* ptrs	to array of	ptrs to	recs */
static PPUI8	papPut;
static PPUI8	papRec0;
static PPUI8	papRec1;
static PPUI8	papDone;

static PPUI8	pBin;					/* ptr to bin (array of	ptrs to	recs) */
static PPPUI8	papBin;					/* ptr to arrays of	ptrs in	bin	*/

static clock_t dwTimeStart;				// clock values
static clock_t dwTimeStop;

static UI32	dwCmp;

/*----------------------------------------------------------------------*/
/*		code															*/
/*----------------------------------------------------------------------*/

static int	GetFile(void);
static void	PutFile(void);
static bool	RecCmp(const void *, const void	*);
static void	InitPointers(void);
static void	MakeBins(void);
static void	SortBins(void);
static void	MakeGroups(void);
static void	MergeGroups(void);

int	main(int argc, char	**argv)
{
using namespace	std;

	vector <PPUI8>	vGet(0);			/* create vectors */
	vector <PPUI8>	vPut(0);
	vector <PUI32>	vGrp(0);
	vector <PPUI8>	vBin(0);
	vector <PPPUI8>	vpBin(0);

	if(argc	!= 3){						/* check arguments */
		cout <<	"hsortp	1.1\n";
		cout <<	"sorts a text file\n";
		cout <<	"usage is:	hsortp <src> <dst>\n";
		cout <<	"<src>			 src file name\n";
		cout <<	"<dst>			 dst file name\n";
		goto exit3;}

	fpSrc =	fopen(argv[1], "rb");		/* open	source file	*/
	if(fpSrc ==	NULL)
		goto exit3;

	fpDst =	fopen(argv[2], "wb");		/* open	dest file */
	if(fpDst ==	NULL)
		goto exit2;

	fseek(fpSrc, 0ul, SEEK_END);		/* get size	of src file	*/
	fgetpos(fpSrc, (fpos_t *)&dwNumBytes);
	fseek(fpSrc, 0ul, SEEK_SET);

	pMem = (PUI8)malloc(dwNumBytes+4);	/* allocate	memory */
	if(pMem	== NULL)
		goto exit1;

	if(GetFile())						/* get the file	*/
		goto exit0;

	vGet.resize(dwNumRec+1);			/* allocate	vectors	*/
	vPut.resize(dwNumRec+1);
	vGrp.resize((dwNumRec+3)/2);
	vBin.resize((dwNumRec+1)*NUMBIN);
	vpBin.resize(NUMBIN);

	papGet	 = (PPUI8)&vGet[0];			/* set ptrs	to vectors */
	papPut	 = (PPUI8)&vPut[0];
	pauGroup = (PUI32)&vGrp[0];
	pBin	 = (PPUI8)&vBin[0];
	papBin	 = (PPPUI8)&vpBin[0];

	InitPointers();						/* set up pointers */

	dwTimeStart	= clock();

	MakeBins();

	SortBins();

	dwTimeStop = clock();

	cout <<	"Number of ticks	 " << dwTimeStop-dwTimeStart <<	"\n";
	cout <<	"Number of compares	 " << dwCmp	<< "\n";

	PutFile();							/* output the file */

exit0:
	free(pMem);							/* release mem */
exit1:
	fclose(fpDst);						/* close dest */
exit2:
	fclose(fpSrc);						/* close source	*/
exit3:
	return(0);
}

/*----------------------------------------------------------------------*/
/*		GetFile()														*/
/*																		*/
/*		Read file in at	pMem+4 (keep on	long boundary).					*/
/*		Prefix each	record with	byte count (replacing ASCLF),			*/
/*		converting to p-string (cnt, data) format starting at pMem+3.	*/
/*		Set	dwNumRec to	# records in file.								*/
/*----------------------------------------------------------------------*/
static int GetFile(void)
{
using namespace	std;

PUI8	pCnt;							/* ptr to record cnt */
PUI8	pRec;							/* ptr to record data */
PUI8	eRec;							/* ptr to record end */
UI32	dwCnt;							/* record count	*/
UI32	i;

/*		read in	file */

	pRec  =	pMem+4;						/* read	in file	*/
	cbSrc =	(UI32)fread(pRec, 1, dwNumBytes, fpSrc);
	pLast  = pRec+cbSrc-1;				/* put ASCLF in	last byte */
	*pLast = ASCLF;
	pCnt  =	pRec-1;						/* init	ptr	to record cnt */

/*		convert	records	to p-strings (cnt, data) */

	i =	0;
	while(pCnt != pLast){
		i++;
		pRec = pCnt+1;
		eRec = (PUI8)memchr((const char	*)pRec,	ASCLF, 0x0100);
		dwCnt =	(UI32)(eRec-pRec);
		if(dwCnt < 0x0001u){
			cout <<	"line "	<< i <<	" is too short \n";
			return(1);}
		if(dwCnt > 0x00ffu){
			cout <<	"line "	<< i <<	" is too long \n";
			return(1);}
		*(pCnt-1) =	0;					/* zero	previous cr	*/
		*(pCnt-0) =	(UI8)dwCnt;			/* set record cnt */
		pCnt  =	eRec;}					/* advance pCnt	*/
	*pCnt =	0;							/* add terminator */
	dwNumRec = i;

	return(0);
}

/*----------------------------------------------------------------------*/
/*		InitPointers													*/
/*																		*/
/*		Set	base pointers based	on number of records.					*/
/*		Create an array	of pointers	to each	record.						*/
/*----------------------------------------------------------------------*/
static void	InitPointers(void)
{
PUI8	pCnt;							/* ptr to rec cnt (and general)	*/
PPUI8	ppRec;							/* ptr to array	of ptrs	to recs	*/

/*		create array of	ptrs to	records	*/

	ppRec =	papGet;						/* ptr to array	of ptrs	to recs	*/
	pCnt = pMem+3;						/* ptr to a	rec	cnt	*/
	while(*pCnt){						/* while more records */
		*ppRec++ = pCnt;				/*	 set a ptr to rec */
		pCnt +=	1+*pCnt;}				/*	 advance ptr to	rec	cnt	*/
	*ppRec = NULL;						/* add terminator to papGet	*/
}

/*----------------------------------------------------------------------*/
/*		PutFile															*/
/*																		*/
/*		Write records from papPut array									*/
/*----------------------------------------------------------------------*/
static void	PutFile(void)
{
PPUI8	ppRec;
PUI8	pRec;
UI32	dwCnt;
UI8		dwSaveCnt;

	ppRec =	papPut;						/* init	ptr	to array */

	while(1){
		if(NULL	== (pRec = *ppRec++)){	/* set ptr to record size */
			break;}
		dwCnt =	(UI32)*pRec++;			/* set dwCnt, pnt to rec data */
		dwSaveCnt =	pRec[dwCnt];		/* save	next rec cnt */
		pRec[dwCnt-1] =	ASCCR;			/* add cr/lf */
		pRec[dwCnt-0] =	ASCLF;			/* write record	*/
		cbDst =	(UI32)fwrite(pRec, 1, dwCnt+1, fpDst);
		pRec[dwCnt]	= dwSaveCnt;}
}

/*----------------------------------------------------------------------*/
/*		RecCmp(void	*elem1,	void *elem2)								*/
/*----------------------------------------------------------------------*/
static bool	RecCmp(const void *elem0, const	void *elem1)
{
	dwCmp += 1;
	if(memcmp(*((PPUI8)elem0)+1, *((PPUI8)elem1)+1,	**((PPUI8)elem0)) <	0)
		return(1);
	else
		return(0);
}

/*----------------------------------------------------------------------*/
/*		MakeBins	seperate ptrs in to	bins							*/
/*----------------------------------------------------------------------*/
static void	MakeBins(void)
{
PPUI8	pTmp;
UI32	i;

	pTmp = pBin;						/* init	papBin */
	for(i =	0; i < NUMBIN; i++){
		papBin[i] =	pTmp;
		pTmp +=	dwNumRec;
	}

	pTmp = papGet;						/* copy	ptrs into pBin */
	while(*pTmp){
		i =	(pTmp[0][1]);
		if(i > NUMBIN-1)
			i =	NUMBIN-1;
		*(papBin[i])++ = *pTmp;
		pTmp++;
	}

	for(i =	0; i < NUMBIN; i++){		/* add terminators to papBin */
		*(papBin[i]) = (PUI8)0;
	}

	pTmp = pBin;						/* reinit papBin */
	for(i =	0; i < NUMBIN; i++){
		papBin[i] =	pTmp;
		pTmp +=	dwNumRec;
	}
}

/*----------------------------------------------------------------------*/
/*		SortBins	sort and write the bins								*/
/*----------------------------------------------------------------------*/
static void	SortBins(void)
{
PPUI8	papPut0;
UI32	i;

	papPut0	= papPut;					/* init	ptr	*/

	for(i =	0; i < NUMBIN; i++){		/* for all bins	*/
		papRec0	= papBin[i];			/*	sort a bin */
		if((*papRec0) == (PUI8)0)
			continue;
		papRec1	= papPut0;
		MakeGroups();
		MergeGroups();
		dwNumRec = pauGroup[0];			/*	number of pts in bin */
		if(papPut0 != papDone)			/*	copy to	papPut if needed */
			memcpy(papPut0,	papDone, dwNumRec*sizeof(PUI8));
		papPut0	+= dwNumRec;			/*	bump output	ptr	*/
	}
	*papPut0 = (PUI8)0;					/* add terminator to list */
}

#if	VARSZGRP

static void	ReversePtrs(PPUI8, UI32);

/*----------------------------------------------------------------------*/
/*		MakeGroups														*/
/*																		*/
/*		Scan file data for ascending and descending	groups of records.	*/
/*		Descending groups of records will have their pointers reversed.	*/
/*		Store a	count into pauGroup[...] for each group.				*/
/*		The	sum	of all group counts	is the number of records in	file.	*/
/*----------------------------------------------------------------------*/
static void	MakeGroups(void)
{
PPUI8	ppRec0;							/* ptr to array	of ptrs	to recs	*/
PPUI8	ppRec1;
PPUI8	ppGrp;							/* ptr to start	of group */
PUI32	pGrp;							/* ptr to array	of counts */
UI32	dwCnt;

	pGrp   = pauGroup;					/* init	ptrs */
	ppGrp =	ppRec1 = papRec0;
	if(*ppRec1 == NULL)					/* stop	if zero	records	*/
		goto done0;

	while(1){

/*		ascending group	*/

		while(1){
			ppRec0 = ppRec1;			/* advance ptrs	to next	rec	*/
			if(*++ppRec1 ==	NULL){		/* stop	if end list	*/
				goto done0;}
			if(RecCmp(ppRec1, ppRec0)){
				break;}}				/* break if	end	ascending group	*/
		dwCnt =	(UI32)(ppRec1-ppGrp);	/* calc	group size */
		if(dwCnt !=	1){					/* if dwCnt	> 1	start new group	*/
			*pGrp++	= dwCnt;			/*	 set # recs	in group */
			ppGrp =	ppRec1;}			/*	 set new group start ptr */

/*		descending group */

		while(1){
			ppRec0 = ppRec1;			/* advance ptrs	to next	rec	*/
			if(*++ppRec1 ==	NULL){		/* stop	if end list	*/
				goto done1;}
			if(!RecCmp(ppRec1, ppRec0)){
				break;}}				/* break if	end	descending group */
		dwCnt =	(UI32)(ppRec1-ppGrp);	/* calc	group size */
		if(dwCnt !=	1){					/* if dwCnt	> 1	start new group	*/
			ReversePtrs(ppGrp, dwCnt);	/*	 reverse the ptrs */
			*pGrp++	= dwCnt;			/*	 set # recs	in group */
			ppGrp =	ppRec1;}}			/*	 set new group start ptr */

done0:									/* end of file,	ascending group	*/
	dwCnt =	(UI32)(ppRec1-ppGrp);		/* calc	group size */
	goto done2;

done1:									/* end of file,	descending group */
	dwCnt =	(UI32)(ppRec1-ppGrp);		/* calc	group size */
	ReversePtrs(ppGrp, dwCnt);			/* reverse the ptrs	*/
done2:
	*pGrp++	= dwCnt;					/* set # recs in group */
	*pGrp =	0;							/* add group terminator	*/
}

/*----------------------------------------------------------------------*/
/*		ReversePtrs		reverse	pointers in	descending group			*/
/*----------------------------------------------------------------------*/
static void	ReversePtrs(PPUI8 ppGrp, UI32 dwCnt)
{
PPUI8	ppSrc;							/* src	ptr	to reverse ptrs	*/
PPUI8	ppDst;							/* dest	ptr	to reverse ptrs	*/

	ppSrc =	ppGrp+dwCnt;				/* set src ptr */
	ppDst =	papRec1;					/* set dest	ptr	*/
	do{*ppDst++	= *--ppSrc;}			/* revsere the ptrs	*/
	while(ppSrc	!= ppGrp);
	memcpy((char *)ppSrc, (char	*)papRec1,
			dwCnt*sizeof(PUI8));
}

#else		/* VARSZGRP	*/

/*----------------------------------------------------------------------*/
/*		MakeGroups														*/
/*																		*/
/*		Scan file data for ascending and descending	pairs of records.	*/
/*		Descending groups of records will have their pointers reversed.	*/
/*		Store a	count into pauGroup[...] for each group.				*/
/*		The	sum	of all group counts	is the number of records in	file.	*/
/*		MakeGroups is essentialy first pass	of MergeGroups				*/
/*----------------------------------------------------------------------*/
static void	MakeGroups(void)
{
PPUI8	ppRec0;							/* ptr to array	of ptrs	to recs	*/
PPUI8	ppRec1;
PUI8	pRec;
PUI32	pGrp;							/* ptr to array	of counts */

	pGrp   = pauGroup;					/* init	ptrs */
	ppRec0 = papRec0;
	ppRec1 = ppRec0+1;
	if(*ppRec0 == NULL)					/* stop	if zero	records	*/
		goto done0;

	while(1){
		if(*ppRec1 == NULL){			/* break if	end	list */
			*pGrp++	= 1;
			break;}
		*pGrp++	= 2;					/*	 set # recs	in group */
		if(RecCmp(ppRec1, ppRec0)){
			pRec	= *ppRec1;			/* if 0	> 1, swap ptrs */
			*ppRec1	= *ppRec0;
			*ppRec0	= pRec;}

		ppRec0 += 2;					/* adv to next pair	*/
		ppRec1 += 2;
		if(*ppRec0 == NULL){			/* break if	end	list */
			break;}}
done0:
	*pGrp =	0;							/* add group terminator	*/
}

#endif		/* VARSZGRP	*/

/*----------------------------------------------------------------------*/
/*		MergeGroups														*/
/*																		*/
/*		Two	groups are merged at a time. One record	pointer	from		*/
/*		each of	a pair of groups is	retrieved. The "smaller" record's	*/
/*		pointer	is appended	to the output array	of pointers. Then the	*/
/*		next pointer is	retrieved from the smaller record's	group.		*/
/*		When the end of	a group	is reached,	the	other group's pointers	*/
/*		are	copied to the output array of pointers.						*/
/*																		*/
/*		Each "pass"	consists of	merging	all	group pairs	from a source	*/
/*		array of pointers to records into a	destination	array. If a		*/
/*		single group is	left over, it's	pointers are copied	over.		*/
/*		After a	pass is	completed, base	pointers to	source and			*/
/*		destination	arrays are swapped.									*/
/*																		*/
/*		Each pass cuts the number of groups	in half. Sorting is			*/
/*		complete when there	is one group left.							*/
/*----------------------------------------------------------------------*/
static void	MergeGroups(void)
{
PUI32	pGrpGet;						/* ptrs	to array of	counts */
PUI32	pGrpPut;
PPUI8	ppGet;							/* current pass	base ptrs */
PPUI8	ppPut;
PPUI8	ppGet0;							/* working ptrs	*/
PPUI8	ppGet1;
PPUI8	ppPut0;
UI32	dwGrpCnt0;						/* group counters */
UI32	dwGrpCnt1;

	ppGet =	papRec0;					/* set up for first	pass */
	ppPut =	papRec1;

/*		start of a pass	*/

	while(pauGroup[1]){					/* while more than 1 group */
		pGrpPut	= pGrpGet =	pauGroup;	/* setup ptrs */
		ppGet1 = ppGet;
		ppPut0 = ppPut;

/*		merge two groups */

		while(1){
			dwGrpCnt0 =	*pGrpGet++;		/* get a set of	counts */
			dwGrpCnt1 =	*pGrpGet++;
			ppGet0	= ppGet1;			/* pnt to group	start */
			ppGet1	= ppGet0+dwGrpCnt0;
			if(!dwGrpCnt0){				/* if group	0 emtpy	*/
				break;}					/*	  end pass */

			(*pGrpPut++) = dwGrpCnt0+dwGrpCnt1;	/* set new group count */

			if(!dwGrpCnt1){				/* if group	1 empty	*/
				while(dwGrpCnt0--){		/*	  copy group 0 */
					*ppPut0++ =	*ppGet0++;}
				break;}					/*	  end pass */

/*		repeatedly copy	the	lower of two record	pointers */
/*		until both groups pointers copied */

			while(1){
				if(!RecCmp(ppGet1, ppGet0)){
					*ppPut0++ =	*ppGet0++;		/* if 0	<= 1, copy 0 */
					if(!--dwGrpCnt0){			/* if end 0, copy 1	*/
						memcpy(ppPut0, ppGet1, dwGrpCnt1*sizeof(PUI8));
						ppGet1 += dwGrpCnt1;
						ppPut0 += dwGrpCnt1;
						break;}}		/* end two group merge */
				else{
					*ppPut0++ =	*ppGet1++;		/* if 0	> 1, copy 1	*/
					if(!--dwGrpCnt1){			/* if end 1, copy 0	*/
						memcpy(ppPut0, ppGet0, dwGrpCnt0*sizeof(PUI8));
						ppGet0 += dwGrpCnt0;
						ppPut0 += dwGrpCnt0;
						break;}}}}		/* end two group merge */

		*pGrpPut = 0;					/* set new group list terminator */

		ppGet0	 = ppGet;				/* swap	ptrs to	array of ptrs */
		ppGet	 = ppPut;
		ppPut	 = ppGet0;}

	papDone	 = ppGet;					/* set ptr to single group */
}
