/*----------------------------------------------------------------------*/
/*      sortp.cpp       text file sort, using sort() or stable_sort()   */
/*                                                                      */
/*      Jeff Reid       13APR22  10:00                                  */
/*----------------------------------------------------------------------*/
/*      Sorts text files: each record ends in CR(0x0d), LF(0x0a).       */
/*                                                                      */
/*      Records are converted into p-strings:                           */
/*          pstring[  0] = c = # bytes in record                        */
/*          pstring[  1] = record data                                  */
/*          pstring[c+1] = 0 (terminator)                               */
/*      An array of pointers to the pstrings is created.                */
/*      The array of pointers is sorted.                                */
/*      The records are output using the array of sorted pointers.      */
/*----------------------------------------------------------------------*/
/*      abbreviations for name prefixes                                 */
/*                                                                      */
/*      a       array of ...                                            */
/*      b       unsigned char (byte)                                    */
/*      cb      count of bytes in ...                                   */
/*      fp      FILE pointer                                            */
/*      p       pointer to ...                                          */
/*      dw      unsigned int (dword)                                    */
/*      qw      unsigned __int64 (qword)                                */
/*      li      LARGE_INTEGER (__int64 structure)                       */
/*----------------------------------------------------------------------*/

/*      get rid of warnings for fopen */
#define _CRT_SECURE_NO_WARNINGS 1

#include <algorithm>
#include <iostream>
#include <vector>

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

/*      ASCCR is carriage return character */
#define ASCCR 0x0d

/*      ASCLF is line feed character */
#define ASCLF 0x0a

typedef unsigned char       UI8;
typedef unsigned char *     PUI8;
typedef unsigned char **    PPUI8;
typedef unsigned int        UI32;
typedef unsigned int  *     PUI32;
typedef unsigned int  **    PPUI32;

typedef bool (* PFCMP)(const void *, const void *);

/*----------------------------------------------------------------------*/
/*      data                                                            */
/*----------------------------------------------------------------------*/

static FILE *   fpSrc;                  /* file desc for source */
static UI32     cbSrc;                  /* count for _dos_read */

static FILE *   fpDst;                  /* file desc dest file */
static UI32     cbDst;                  /* count for _dos_write */

static UI32     dwNumBytes;             /* # bytes in file */
static UI32     dwNumRec;               /* # of records in file */
static PUI8     pMem;                   /* allocated memory buffer */
static PUI8     pLast;                  /* ptr to last UI8 in file */

static PFCMP    pfCmp;
static PPUI8    papRec;                 /* ptrs to array of ptrs to recs */

static clock_t dwTimeStart;             // clock values
static clock_t dwTimeStop;

static UI32 dwCmp;

/*----------------------------------------------------------------------*/
/*      code                                                            */
/*----------------------------------------------------------------------*/

static int  GetFile(void);
static void InitPointers(void);
static void PutFile(void);
static bool RecCmp(const void *, const void *);

int main(int argc, char **argv)
{
using namespace std;

	pfCmp = RecCmp;                     /* set ptr to compare function */
	vector <PPUI8> vPtr(0);             /* create vector */

	if(argc != 3){                      /* check arguments */
		cout << "sortp 1.1\n";
		cout << "sorts a text file\n";
		cout << "usage is:  sortp <src> <dst>\n";
		cout << "<src>           src file name\n";
		cout << "<dst>           dst file name\n";
		goto exit3;}

	fpSrc = fopen(argv[1], "rb");       /* open source file */
	if(fpSrc == NULL)
		goto exit3;

	fpDst = fopen(argv[2], "wb");       /* open dest file */
	if(fpDst == NULL)
		goto exit2;

	fseek(fpSrc, 0ul, SEEK_END);        /* get size of src file */
	fgetpos(fpSrc, (fpos_t *)&dwNumBytes);
	fseek(fpSrc, 0ul, SEEK_SET);

	pMem = (PUI8)malloc(dwNumBytes+4);  /* allocate memory */
	if(pMem == NULL)
		goto exit1;

	if(GetFile())                       /* get the file */
		goto exit0;

	vPtr.resize(dwNumRec);              /* allocate vector */
	papRec = (PPUI8)&vPtr[0];
	InitPointers();                     /* init pointers */

	dwTimeStart = clock();

	stable_sort(vPtr.begin(), vPtr.end(), pfCmp);

	dwTimeStop = clock();

	cout << "Number of ticks     " << dwTimeStop-dwTimeStart << "\n";
	cout << "Number of compares  " << dwCmp << "\n";

	PutFile();                          /* output the file */

exit0:
	free(pMem);                         /* release mem */
exit1:
	fclose(fpDst);                      /* close dest */
exit2:
	fclose(fpSrc);                      /* close source */
exit3:
	return(0);
}

/*----------------------------------------------------------------------*/
/*      GetFile()                                                       */
/*                                                                      */
/*      Read file in at pMem+4 (keep on long boundary).                 */
/*      Prefix each record with byte count (replacing ASCLF),           */
/*      converting to p-string (cnt, data) format starting at pMem+3.   */
/*      Set dwNumRec to # records in file.                              */
/*----------------------------------------------------------------------*/
static int GetFile(void)
{
using namespace std;

PUI8    pCnt;                           /* ptr to record cnt */
PUI8    pRec;                           /* ptr to record data */
PUI8    eRec;                           /* ptr to record end */
UI32    dwCnt;                          /* record count */
UI32    i;

/*      read in file */

	pRec  = pMem+4;                     /* read in file */
	cbSrc = (UI32)fread(pRec, 1, dwNumBytes, fpSrc);
	pLast  = pRec+cbSrc-1;              /* put ASCLF in last byte */
	*pLast = ASCLF;
	pCnt  = pRec-1;                     /* init ptr to record cnt */

/*      convert records to p-strings (cnt, data) */

	i = 0;
	while(pCnt != pLast){
		i++;
		pRec = pCnt+1;
		eRec = (PUI8)memchr((const char *)pRec, ASCLF, 0x0100);
		dwCnt = (UI32)(eRec-pRec);
		if(dwCnt < 0x0001u){
			cout << "line " << i << " is too short \n";
			return(1);}
		if(dwCnt > 0x00ffu){
			cout << "line " << i << " is too long \n";
			return(1);}
		*(pCnt-1) = 0;                  /* zero previous cr */
		*(pCnt-0) = (UI8)dwCnt;         /* set record cnt */
		pCnt  = eRec;}                  /* advance pCnt */
	*pCnt = 0;                          /* add terminator */
	dwNumRec = i;

	return(0);
}

/*----------------------------------------------------------------------*/
/*      InitPointers                                                    */
/*                                                                      */
/*      Set base pointers based on number of records.                   */
/*      Create an array of pointers to each record.                     */
/*----------------------------------------------------------------------*/
static void InitPointers()
{
PUI8    pCnt;                           /* ptr to rec cnt (and general) */
PPUI8   ppRec;                          /* ptr to array of ptrs to recs */

/*      create array of ptrs to records */

	ppRec = papRec;                     /* ptr to array of ptrs to recs */
	pCnt = pMem+3;                      /* ptr to a rec cnt */
	while(*pCnt){                       /* while more records */
		*ppRec++ = pCnt;                /*   set a ptr to rec */
		pCnt += 1+*pCnt;}               /*   advance ptr to rec cnt */
}

/*----------------------------------------------------------------------*/
/*      PutFile                                                         */
/*                                                                      */
/*      Write records from papRec array                                 */
/*----------------------------------------------------------------------*/
static void PutFile()
{
PPUI8   ppRec;
PUI8    pRec;
UI32    dwCnt;
UI8     dwSaveCnt;

	ppRec = papRec;                     /* init ptr to array */

	while(1){
		if(NULL == (pRec = *ppRec++)){  /* set ptr to record count */
			break;}
		dwCnt = (UI32)*pRec++;          /* set dwCnt, pnt to rec data */
		dwSaveCnt = pRec[dwCnt];        /* save next rec cnt */
		pRec[dwCnt-1] = ASCCR;          /* add cr/lf */
		pRec[dwCnt-0] = ASCLF;          /* write record */
		cbDst = (UI32)fwrite(pRec, 1, dwCnt+1, fpDst);
		pRec[dwCnt] = dwSaveCnt;}
}

/*----------------------------------------------------------------------*/
/*      RecCmp(void *elem1, void *elem2)                                */
/*----------------------------------------------------------------------*/
static bool RecCmp(const void *elem0, const void *elem1)
{
	dwCmp += 1;
	if(memcmp(((PUI8)elem0)+1, ((PUI8)elem1)+1, *((PUI8)elem0)) < 0)
		return(1);
	else
		return(0);
}

